(*-------------------------------------------------------------------------*)
(*                                                                         *)
(*  Amiga Oberon Interface Module: BasicTypes         Date: 02-Nov-92      *)
(*                                                                         *)
(*    1992 by Fridtjof Siebert                                            *)
(*                                                                         *)
(*-------------------------------------------------------------------------*)

MODULE BasicTypes;

(*
 * This module defines some basic object types and the basic features
 * applicable to them.
 *
 * Anytime a new class is defined, it should be taken into consideration
 * that it may be interpreted as a descandant of one of the types defined
 * here. If this is the case, it should inherit the features defined here
 * and redefine them accordingly.
 *
 * Any class that, by its nature, can't be seen as a descandant of one of
 * the classes defined here, should at least be declared as a descandant
 * of class ANY.
 *
 *)


TYPE
(*
 * You should define every new class as a descendant
 * of ANY to gain compatibility.
 *)

  ANY * = POINTER TO ANYDesc;
  ANYDesc * = RECORD END;

(*
 * COLLECTION should be the base class of classes that implement a
 * collection of some kind of elements, such as lists or arrays.
 * Note that the routine features of COLLECTION fail unless they
 * are redefined.
 *)

  COLLECTION * = POINTER TO COLLECTIONDesc;
  COLLECTIONDesc * = RECORD (ANYDesc) END;

  DoProc * = PROCEDURE(x,par: ANY);


(*
 * Mathematical objects that may be compared and that build up a
 * hirarchy.
 *)
  COMPAREABLE * = POINTER TO COMPAREABLEDesc;
  COMPAREABLEDesc * = RECORD (ANYDesc) END;


(*
 * A structure building a mathematical group shoulb be a heir
 * of the following type:
 *)

  GROUP * = POINTER TO GROUPDesc;
  GROUPDesc * = RECORD (COMPAREABLEDesc) END;


(*
 * A sructur building a mathematical ring should be a heir of the
 * following type:
 *)

  RING * = POINTER TO RINGDesc;
  RINGDesc * = RECORD (GROUPDesc) END;


(*
 * A sructur building a mathematical field should be implemented as
 * a heir of FIELD.
 *)

  FIELD * = POINTER TO FIELDDesc;
  FIELDDesc * = RECORD (RINGDesc) END;


(*
 * General Type for dynamically allocated strings with an arbitraty length:
 *
 *)

  DynString * = POINTER TO ARRAY OF CHAR;


(*
 * routines for COLLECTION:
 *
 *)


PROCEDURE (c: COLLECTION) Add        * (x: ANY);             BEGIN HALT(20) END Add;
(* add x to c.
 * depending on the implementation it is added at some point to the
 * collection.
 *)

PROCEDURE (c: COLLECTION) Remove     * (x: ANY);              BEGIN HALT(20) END Remove;
(* removes x from c.
 *)

PROCEDURE (c: COLLECTION) nbElements * (): LONGINT;           BEGIN HALT(20) END nbElements;
(* returns the number of elements within c.
 *)

PROCEDURE (c: COLLECTION) isEmpty    * (): BOOLEAN;           
BEGIN 
  RETURN c.nbElements()=0 
END isEmpty;
(* TRUE if c is empty
 *)

PROCEDURE (c: COLLECTION) Do         * (p: DoProc; par: ANY); BEGIN HALT(20) END Do;
(* calls p(x,par) for every element x stored within c.
 * par passes some additional information to p. par is not touched by Do.
 *)


(*
 * routines of COMPAREABLE:
 *
 * These routines should be redefined and to implement comparisons. Note that
 * it's usually sufficient to redefine COMPAREABLE.Less and COMPAREABLE.Equal.
 *
 *)

PROCEDURE (a: COMPAREABLE) Compare       * (b: COMPAREABLE): LONGINT; (* result = 0 <=> a = b *)
                                                                      (* result < 0 <=> a < b *)
                                                                      (* result > 0 <=> a > b *)
BEGIN HALT(20) END Compare;
PROCEDURE (a: COMPAREABLE) Equal         * (b: COMPAREABLE): BOOLEAN; (* a =  b *) BEGIN RETURN a.Compare(b)= 0 END Equal;
PROCEDURE (a: COMPAREABLE) Less          * (b: COMPAREABLE): BOOLEAN; (* a <  b *) BEGIN RETURN a.Compare(b)< 0 END Less;
PROCEDURE (a: COMPAREABLE) LessOrEqual   * (b: COMPAREABLE): BOOLEAN; (* a <= b *) BEGIN RETURN a.Compare(b)<=0 END LessOrEqual;
PROCEDURE (a: COMPAREABLE) Higher        * (b: COMPAREABLE): BOOLEAN; (* a >  b *) BEGIN RETURN a.Compare(b)> 0 END Higher;
PROCEDURE (a: COMPAREABLE) HigherOrEqual * (b: COMPAREABLE): BOOLEAN; (* a >= b *) BEGIN RETURN a.Compare(b)>=0 END HigherOrEqual;


(*
 * routines of GROUP:
 *
 * The following methods are applicable to groups:
 *
 *   Add: A (possibly cummutative) operation on two elements.
 *   Neg: Build complement of an Element of the group.
 *   Sub: Add comlement of an element to another element.
 *   Norm: returns a norm (a positive real number) for the specified element
 *
 * All of them should allocate a new object and return a reference to it.
 *
 * The following axiomes should be satisfied:
 *
 * 1) Add must be associative <==> (a Add b) Add c = a Add (b Add c)
 * 2) There must be a neutral element z so that x Add z = x for all x
 * 3) There must be a inverse element x' for any x so that x Add x' = z
 * 4) x.Norm() = 0 <==> x.Equal(z)
 *
 *)

PROCEDURE (m: GROUP) Add  * (n: GROUP): GROUP;    BEGIN HALT(20) END Add;
PROCEDURE (m: GROUP) Neg  * (        ): GROUP;    BEGIN HALT(20) END Neg;
PROCEDURE (m: GROUP) Sub  * (n: GROUP): GROUP;    BEGIN RETURN m.Add(n.Neg()) END Sub;
PROCEDURE (m: GROUP) Norm * (        ): LONGREAL; BEGIN HALT(20) END Norm;


(*
 * routines of RING:
 *
 *  due to inheritance, all routines of class GROUP plus
 *   Mul : A (possibly commutative) operation on two elements
 *   Sqr : a Mul a
 *
 * In addition to the axioms of GROUP, the following axiomes should be
 * satisfied:
 *
 * 1) Mul must associative <==>  (a Mul b) Mul c = a Mul (b Mul c)
 * 2) a Mul (b Add c) = (a Mul b) Add (a Mul c) for all a,b,c
 * 3) (b Add c) Mul a = (b Mul a) Add (c Mul a) for all a,b,c
 *
 * The routines should be redifined and to implement the basic mathematic
 * operations for a special of mathematical object.
 *
 * All of them should allocate a new object and return a reference to it.
 *
 *)

PROCEDURE (m: RING) Mul  * (n: RING): RING;     BEGIN HALT(20) END Mul;
PROCEDURE (m: RING) Sqr  * (       ): RING;     BEGIN RETURN m.Mul(m); END Sqr;


(*
 * routines of FIELD:
 *
 *  due to inheritance, all routines of class RING plus
 *   Inv: return inverse element. x Mul (Inv x) = z
 *   InvAllowed: return FALSE if and only if a.Inv will succeed.
 *        x.Inv should cause a division by zero error if and only if
 *        ~ x.InvAllowed()
 *   Div: x Div y = x Mul (Inv y).
 *
 * In addition to the axioms of RING, the following axiomes should be
 * satisfied:
 *
 * 1) a Add b = b Add a
 * 2) a Mul b = b Mul a
 * 3) There must be an element e so that x Mul e = x for all x
 * 4) There must be an inverse element x' for any x#z, so that x Mul x' = e
 *    (z = neutral element of Add)
 *
 * The routines should be redifined and to implement the basic mathematic
 * operations for a special of mathematical object.
 *
 * All of them should allocate a new object and return a reference to it.
 *
 *)

PROCEDURE (m: FIELD) Inv * (       ): FIELD; BEGIN HALT(20) END Inv;
PROCEDURE (m: FIELD) InvAllowed * (): BOOLEAN; BEGIN HALT(20) END InvAllowed;
PROCEDURE (m: FIELD) Div * (n: FIELD): FIELD;
VAR
  o: RING;
BEGIN
  o := m.Mul(n.Inv());
  RETURN o(FIELD);
END Div;


END BasicTypes.








